import concurrent.futures
import mmap
import pathlib
import re
from typing import Union

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import requests
import tqdm

from tacoreader import compile_utils


def compile(
    dataframe: pd.DataFrame,
    output: Union[str, pathlib.Path],
    chunk_size_iter: str = "100MB",
    nworkers: int = 4,
    overwrite: bool = True,
    quiet: bool = False,
) -> pathlib.Path:
    """Select a subset of a Tortilla or TACO file and write a new "small"
    Tortilla file. If you want to convert this new file to a TACO file,
    you can use the `tacotoolbox.tortilla2taco` function.

    Args:
        metadata (pd.DataFrame): A subset of the Tortilla file.
        output_folder (Union[str, pathlib.Path]): The folder where the Tortilla file
            will be saved. If the folder does not exist, it will be created.
        chunk_size_iter (int, optional): The writting chunk size. By default,
            it is 100MB. Faster computers can use a larger chunk size.
        nworkers (int, optional): The number of workers to use when writing
            the tortilla. Defaults to 4.
        overwrite (bool, optional): If True, the function overwrites the file if it
            already exists. By default, it is True.
        quiet (bool, optional): If True, the function does not print any
            message. By default, it is False.
    Returns:
        pathlib.Path: The path to the new Tortilla file.
    """
    # From human-readable to bytes
    chunk_size_iter: int = compile_utils.human2bytes(chunk_size_iter)

    # If the folder does not exist, create it
    output = pathlib.Path(output)
    output.parent.mkdir(parents=True, exist_ok=True)

    # Check if the file already exists
    if output.exists() and overwrite:
        output.unlink()

    # Remove the index from the previous dataset
    dataframe = dataframe.copy()
    dataframe.sort_values("tortilla:offset", inplace=True)
    dataframe.reset_index(drop=True, inplace=True)

    # Compile your tortilla
    mode = dataframe["internal:mode"].iloc[0]
    if mode == "local":
        compile_local(dataframe, output, chunk_size_iter, nworkers, quiet)
    elif mode == "online":
        compile_online(dataframe, output, chunk_size_iter, nworkers, quiet)
    else:
        raise ValueError(f"Unknown mode: {mode}")

    return output


def compile_local(
    dataframe: pd.DataFrame,
    output: str,
    chunk_size_iter: int,
    nworkers: int,
    quiet: bool,
) -> pathlib.Path:
    """Prepare a subset of a local Tortilla file and write it to a new file.

    Args:
        dataframe (pd.DataFrame): The dataframe of a Tortilla file.
        output (str): The path to the Tortilla file.
        chunk_size_iter (int): The size of the chunks to use when writing
            the tortilla.
        nworkers (int): The number of workers to use when writing the tortilla.
        quiet (bool): Whether to suppress the progress bar.

    Returns:
        pathlib.Path: The path to the new Tortilla file.
    """

    # Estimate the new offset
    dataframe.loc[:, "tortilla:new_offset"] = (
        dataframe["tortilla:length"].shift(1, fill_value=0).cumsum() + 200
    )

    # Create the new FOOTER
    # Remove the columns generated on-the-fly by the load function (internal fields)
    new_footer = dataframe.copy()
    new_footer.drop(
        columns=[
            "geometry",
            "internal:mode",
            "internal:subfile",
            "tortilla:offset",
        ],
        errors="ignore",
        inplace=True,
    )
    new_footer.rename(columns={"tortilla:new_offset": "tortilla:offset"}, inplace=True)

    # Create an in-memory Parquet file with BufferOutputStream
    with pa.BufferOutputStream() as sink:
        pq.write_table(
            pa.Table.from_pandas(new_footer),
            sink,
            compression="zstd",  # Highly efficient codec
            compression_level=22,  # Maximum compression for Zstandard
            use_dictionary=False,  # Optimizes for repeated values
        )
        # return a blob of the in-memory Parquet file as bytes
        # Obtain the FOOTER metadata
        FOOTER: bytes = sink.getvalue().to_pybytes()

    # Calculate the bytes of the data blob (DATA)
    bytes_counter: int = (
        dataframe.iloc[-1]["tortilla:new_offset"]
        + dataframe.iloc[-1]["tortilla:length"]
    )

    # Prepare the static bytes
    MB: bytes = b"#y"
    FL: bytes = len(FOOTER).to_bytes(8, "little")
    FO: bytes = int(bytes_counter).to_bytes(8, "little")
    DP: bytes = int(1).to_bytes(8, "little")

    # Create the tortilla file (empty)
    with open(output, "wb") as f:
        f.truncate(bytes_counter + len(FOOTER))

    # Define the function to write into the main file
    def write_file(file, old_offset, length, new_offset):
        """read the file in chunks"""
        with open(file, "rb") as g:
            g.seek(old_offset)
            while True:
                # Iterate over the file in chunks until the length is 0
                if chunk_size_iter > length:
                    chunk = g.read(length)
                    length = 0
                else:
                    chunk = g.read(chunk_size_iter)
                    length -= chunk_size_iter

                # Write the chunk into the mmap
                mm[new_offset : (new_offset + len(chunk))] = chunk
                new_offset += len(chunk)

                if length == 0:
                    break

    # Cook the tortilla 🫓
    with open(output, "r+b") as f:
        with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_WRITE) as mm:
            # Write the magic number
            mm[:2] = MB

            # Write the FOOTER offset
            mm[2:10] = FO

            # Write the FOOTER length
            mm[10:18] = FL

            # Write the DATA PARTITIONS
            mm[18:26] = DP

            # Write the free space
            mm[26:200] = b"\0" * 174

            # Write the DATA
            message = compile_utils.tortilla_message()
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=nworkers
            ) as executor:
                futures = []
                for _, item in dataframe.iterrows():
                    old_offset = item["tortilla:offset"]
                    new_offset = item["tortilla:new_offset"]
                    length = item["tortilla:length"]
                    file = dataframe["internal:subfile"].iloc[0].split(",")[-1]
                    futures.append(
                        executor.submit(
                            write_file, file, old_offset, length, new_offset
                        )
                    )

                # Wait for all futures to complete
                if not quiet:
                    list(
                        tqdm.tqdm(
                            concurrent.futures.as_completed(futures),
                            total=len(futures),
                            desc=message,
                            unit="file",
                        )
                    )
                else:
                    concurrent.futures.wait(futures)

            # Write the FOOTER
            mm[bytes_counter : (bytes_counter + len(FOOTER))] = FOOTER

    return pathlib.Path(output)


def compile_online(
    dataframe: pd.DataFrame,
    output: str,
    chunk_size_iter: int,
    nworkers: int,
    quiet: bool,
) -> pathlib.Path:
    """Prepare a subset of an online Tortilla or TACO file and write it to
    a new local file. If you want to convert this new file to a TACO file,
    you can use the `tacotoolbox.tortilla2taco` function.

    Args:
        metadata (pd.DataFrame): The metadata of the Tortilla file.
        output (str): The path to the Tortilla file.
        chunk_size_iter (int, optional): The size of the chunks to use
            when writing the tortilla.
        quiet (bool): Whether to suppress the progress bar.

    Returns:
        pathlib.Path: The path to the new Tortilla file.
    """

    # Get the URL of the file
    url_pattern = r"(ftp|https?)://[^\s,]+"
    url = re.search(url_pattern, dataframe["internal:subfile"].iloc[0]).group(0)

    # Calculate the new offsets
    dataframe["tortilla:new_offset"] = (
        dataframe["tortilla:length"].shift(1, fill_value=0).cumsum() + 200
    )

    # Create the new FOOTER
    # Remove the columns generated on-the-fly by the load function
    new_footer = dataframe.copy()
    new_footer.drop(
        columns=[
            "geometry",
            "internal:mode",
            "internal:subfile",
            "tortilla:offset",
        ],
        errors="ignore",
        inplace=True,
    )
    new_footer.rename(columns={"tortilla:new_offset": "tortilla:offset"}, inplace=True)

    # Create an in-memory Parquet file with BufferOutputStream
    with pa.BufferOutputStream() as sink:
        pq.write_table(
            pa.Table.from_pandas(new_footer),
            sink,
            compression="zstd",  # Highly efficient codec
            compression_level=22,  # Maximum compression for Zstandard
            use_dictionary=False,  # Optimizes for repeated values
        )
        # return a blob of the in-memory Parquet file as bytes
        # Obtain the FOOTER metadata
        FOOTER: bytes = sink.getvalue().to_pybytes()

    # Calculate the total size of the data
    bytes_counter: int = (
        dataframe.iloc[-1]["tortilla:new_offset"] + dataframe.iloc[-1]["tortilla:length"]
    )

    # Prepare static bytes
    MB: bytes = b"#y"
    FL: bytes = len(FOOTER).to_bytes(8, "little")
    FO: bytes = int(bytes_counter).to_bytes(8, "little")
    DP: bytes = int(1).to_bytes(8, "little")


    # Define the function to write into the main file
    def write_file_url(url, header, new_offset):
        """read the url in chunks"""
        try:
            with requests.get(url, headers=header, stream=True, timeout=10) as response:
                response.raise_for_status()
                # Write the downloaded data in chunks and update the progress bar
                for chunk in response.iter_content(chunk_size=chunk_size_iter):
                    if chunk:  # Filter out keep-alive new chunks
                        mm[new_offset : (new_offset + len(chunk))] = chunk
                        new_offset += len(chunk)
        except requests.exceptions.RequestException as e:
            print("ddd")
            raise requests.exceptions.RequestException(
                f"Error downloading {url} with header {header}: {e}"
            )
                                                                             
    # Create the tortilla file (empty)
    with open(output, "wb") as f:
        f.truncate(bytes_counter + len(FOOTER))

    # Cook the tortilla 🫓
    with open(output, "r+b") as f:
        with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_WRITE) as mm:
            # Write the magic number
            mm[:2] = MB

            # Write the FOOTER offset
            mm[2:10] = FO

            # Write the FOOTER length
            mm[10:18] = FL

            # Write the DATA PARTITIONS
            mm[18:26] = DP

            # Write the free space
            mm[26:200] = b"\0" * 174

            # Write the DATA
            message = compile_utils.tortilla_message()
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=nworkers
            ) as executor:
                futures = []
                for _, item in dataframe.iterrows():
                    # Get the offset, length, and URL
                    new_offset = item["tortilla:new_offset"]
                    url = dataframe["internal:subfile"].iloc[0].split(",")[-1].replace("/vsicurl/", "")
                    
                    # Create the header
                    next_start = item["tortilla:offset"]
                    next_end = next_start + item["tortilla:length"]
                    header = {"Range": f"bytes={next_start}-{next_end - 1}"}

                    futures.append(
                        executor.submit(
                            write_file_url, url, header, new_offset
                        )
                    )

                # Wait for all futures to complete
                if not quiet:
                    list(
                        tqdm.tqdm(
                            concurrent.futures.as_completed(futures),
                            total=len(futures),
                            desc=message,
                            unit="file",
                        )
                    )
                else:
                    concurrent.futures.wait(futures)

            # Write the FOOTER
            mm[bytes_counter : (bytes_counter + len(FOOTER))] = FOOTER
        
    return pathlib.Path(output)


import pathlib
import re
from typing import List, Union

import geopandas as gpd
import pandas as pd
import requests

from tacoreader import datasets, load_local, load_remote, load_utils


def load(file: Union[str, pathlib.Path, List[pathlib.Path], List[str]]) -> pd.DataFrame:
    """Load the dataframe of a tortilla file.

    Args:
        file (Union[str, pathlib.Path, List]): The path of
            the tortilla file. If the file is split into multiple
            parts, a list of paths is accepted. Also, multiple
            parts can be read by putting a asterisk (*) at the end
            of the file name. For example, "file*.tortilla". In this
            case, the function will create a list will all the partitions
            before the reading process.

    Returns:
        pd.DataFrame: The dataframe of the tortilla file.
    """


    if isinstance(file, list):
        if load_utils.is_valid_url(file):
            dataframe = load_remote.remote_files2dataframe(file)
        else:
            dataframe = load_local.local_files2dataframe(file)
    elif isinstance(file, (str, pathlib.Path)):
        # Special case for the taco foundation
        if file.startswith("tacofoundation:"):
            # Check if the dataset is inside the taco foundation
            if file in datasets.datasets.keys():
                file = datasets.datasets[file]
            else:
                raise ValueError(
                    "The dataset is not available in the Taco Foundation repository."
                )

        # Transform our snippet into a list of files
        # If it is not a snippet, it will return the same file
        file = load_utils.snippet2files(file=file)

        if load_utils.is_valid_url(file):
            dataframe = load_remote.remote_file2dataframe(file)
        else:
            dataframe = load_local.local_file2dataframe(file)
    else:
        raise ValueError("Invalid file type. Must be a list, string or pathlib.Path.")

    # Clean up the dataframe
    dataframe = load_utils.sort_columns_add_geometry(dataframe)

    return TortillaDataFrame(dataframe)


def lazy_load(offset: int, file: Union[str, pathlib.Path]) -> pd.DataFrame:
    """Lazy load a tortilla file.

    Useful for datasets that have tortillas as samples (tortillas inside tortillas).
    The offset is used to read a specific part of the main tortilla file.

    Args:
        offset (int): The byte offset where the reading process will start.
        file (Union[str, pathlib.Path]): The path tot the main tortilla file.

    Returns:
        pd.DataFrame: The dataframe of the tortilla file.
    """

    if load_utils.is_valid_url(file):
        dataframe = load_remote.remote_lazyfile2dataframe(offset, file)
    else:
        dataframe = load_local.local_lazyfile2dataframe(offset, file)

    # Clean up the dataframe
    dataframe = load_utils.sort_columns_add_geometry(dataframe)

    return TortillaDataFrame(dataframe)


def load_metadata(
    file: Union[str, pathlib.Path, List[pathlib.Path], List[str]]
) -> dict:
    """Load the metadata of a tortilla or taco file.

    Args:
        file (Union[str, pathlib.Path, List]): The path of
            the taco file. If the file is split into multiple
            parts, a list of paths is accepted. Also, multiple
            parts can be read by putting a asterisk (*) at the end
            of the file name. For example, "file*.tortilla". In this
            case, the function will create a list will all the partitions
            before the reading process.

    Returns:
        dict: The metadata of the taco file.
    """
    # Transform our snippet into a list of files
    # If it is not a snippet, it will return the same file
    file = load_utils.snippet2files(file=file)

    # Load the metadata
    if isinstance(file, list):
        if load_utils.is_valid_url(file):
            metadata = load_remote.remote_files2metadata(file)
        else:
            metadata = load_local.local_files2metadata(file)
    elif isinstance(file, (str, pathlib.Path)):
        if load_utils.is_valid_url(file):
            metadata = load_remote.remote_file2metadata(file)
        else:
            metadata = load_local.local_file2metadata(file)
    else:
        raise ValueError("Invalid file type. Must be a list, string or pathlib.Path.")

    return metadata


class TortillaDataFrame(gpd.GeoDataFrame):

    @property
    def _constructor(self):
        return TortillaDataFrame

    def __finalize__(self, other, method=None, **kwargs):
        # Propagate custom metadata
        if isinstance(other, TortillaDataFrame):
            for name in self._metadata:
                setattr(self, name, getattr(other, name, None))
        return self

    @staticmethod
    def get_internal_path(row):
        """
        Extract offset, length, and path from a row's internal subfile information.
        """
        pattern: re.Pattern = re.compile(r"/vsisubfile/(\d+)_(\d+),(.+)")
        offset, length, path = pattern.match(row["internal:subfile"]).groups()

        # Adjust path for curl files
        if path.startswith("/vsicurl/"):
            path = path[9:]

        return int(offset), int(length), path

    def read(self, idx):
        """
        Read data based on the row's tortilla:file_format.
        """
        row = self.iloc[idx]
        if row["tortilla:file_format"] == "TORTILLA":
            offset, length, path = self.get_internal_path(row)
            return lazy_load(row["tortilla:offset"], path)
        elif row["tortilla:file_format"] == "BYTES":
            # Obtain the offset, length and internal path
            offset, length, path = self.get_internal_path(row)

            # Fetch the bytes
            if load_utils.is_valid_url(path):
                headers = {"Range": f"bytes={offset}-{offset + length - 1}"}
                response: requests.Response = requests.get(path, headers=headers)
                return response.content
            else:
                with open(path, "rb") as f:
                    f.seek(int(offset))
                    return f.read(int(length))
        else:
            return row["internal:subfile"]



import json
import os
import pathlib
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import List, Union

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import requests


def remote_file2dataframe(file: str) -> pd.DataFrame:
    """Fetch and parse the metadata of a tortilla file."""
    headers = {"Range": "bytes=0-17"}
    response = requests.get(file, headers=headers)
    static_bytes = response.content

    # Extract metadata
    MB, FO, FL = static_bytes[:2], static_bytes[2:10], static_bytes[10:18]
    if MB not in {b"#y", b"WX"}:
        if requests.head(file).status_code == 404:
            raise FileNotFoundError(f"File not found: {file}")
        raise ValueError("Invalid file type: must be either a Tortilla 🫓 or a TACO 🌮")
    footer_offset = int.from_bytes(FO, "little")
    footer_length = int.from_bytes(FL, "little")

    # Fetch the footer
    headers = {"Range": f"bytes={footer_offset}-{footer_offset + footer_length - 1}"}
    response = requests.get(file, headers=headers)
    dataframe = pq.read_table(pa.BufferReader(response.content)).to_pandas()

    # Add auxiliary columns
    dataframe["internal:mode"] = "online"
    dataframe["internal:subfile"] = dataframe.apply(
        lambda row: f"/vsisubfile/{row['tortilla:offset']}_{row['tortilla:length']},/vsicurl/{file}",
        axis=1,
    )
    return dataframe


def remote_files2dataframe(files: List[str]) -> pd.DataFrame:
    """Read metadata from multiple tortilla files in parallel."""
    max_workers = len(files) if len(files) < os.cpu_count() else os.cpu_count()
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(remote_file2dataframe, url) for url in files]
        results = []
        for future in as_completed(futures):
            result = future.result()
            if result is not None:
                results.append(result)
    return pd.concat(results, ignore_index=True)


def remote_lazyfile2dataframe(
    offset: int, file: Union[str, pathlib.Path]
) -> pd.DataFrame:
    """Read the dataframe of tortilla file that is a subfile
    of a larger file.

    Args:
        offset (int): The offset of the subfile.
        file (Union[str, pathlib.Path]): A local path pointing to the
            main tortilla file.

    Returns:
        pd.DataFrame: The dataframe of the tortilla file.
    """

    # Fetch the first 8 bytes of the file
    initb, endb = offset, offset + 18
    headers = {"Range": f"bytes={initb}-{endb - 1}"}
    response: requests.Response = requests.get(file, headers=headers)
    static_bytes: bytes = response.content

    # SPLIT the static bytes
    MB: bytes = static_bytes[:2]
    FO: bytes = static_bytes[2:10]
    FL: bytes = static_bytes[10:18]

    # Check if the file is a tortilla
    if MB not in {b"#y", b"WX"}:
        if requests.head(file).status_code == 404:
            raise FileNotFoundError(f"File not found: {file}")
        raise ValueError("Invalid file type: must be either a Tortilla 🫓 or a TACO 🌮")

    # Interpret the bytes as a little-endian integer
    footer_offset: int = int.from_bytes(FO, "little") + offset
    footer_length: int = int.from_bytes(FL, "little")

    # Fetch the footer
    headers = {"Range": f"bytes={footer_offset}-{footer_offset + footer_length - 1}"}
    with requests.get(file, headers=headers) as response:
        # Interpret the response as a parquet table
        dataframe = pq.read_table(pa.BufferReader(response.content)).to_pandas()

    # Fix the offset
    dataframe["tortilla:offset"] = dataframe["tortilla:offset"] + offset

    # Add the file format and mode
    dataframe["internal:mode"] = "online"
    dataframe["internal:subfile"] = dataframe.apply(
        lambda row: f"/vsisubfile/{row['tortilla:offset']}_{row['tortilla:length']},/vsicurl/{file}",
        axis=1,
    )

    return dataframe


def remote_file2metadata(file: str) -> dict:
    """Read the metadata of a taco file given a URL. The
        server must support HTTP Range requests.

    Args:
        file (str): A URL pointing to the taco file.

    Returns:
        dict: The metadata of the taco file.
    """
    # Fetch the first 8 bytes of the file
    headers = {"Range": "bytes=0-41"}
    response: requests.Response = requests.get(file, headers=headers)
    static_bytes: bytes = response.content

    # SPLIT the static bytes
    MB: bytes = static_bytes[:2]
    CO: int = int.from_bytes(static_bytes[26:34], "little")
    CL: int = int.from_bytes(static_bytes[34:42], "little")

    # Check if the file is a tortilla
    if MB not in {b"#y", b"WX"}:
        if requests.head(file).status_code == 404:
            raise FileNotFoundError(f"File not found: {file}")
        raise ValueError("Invalid file type: must be either a Tortilla 🫓 or a TACO 🌮")

    # Read the Collection (JSON UTF-8 encoded)
    headers = {"Range": f"bytes={CO}-{CO + CL - 1}"}
    collection: dict = json.loads(requests.get(file, headers=headers).content.decode())

    return collection


def remote_files2metadata(files: List[str]) -> dict:
    """Read the metadata of taco files given a set of URLs. The server
        must support HTTP Range requests.

    Args:
        files (List[str]): A list of URLs pointing to the
            taco files.

    Returns:
        dict: The metadata of the taco file.
    """
    return remote_file2metadata(files[0])



from tacoreader.load import load, load_metadata, TortillaDataFrame
from tacoreader.compile import compile
import pandas as pd
import geopandas as gpd

__all__ = ["load", "compile", "load_metadata"]
__version__ = "0.4.7"

# TODO: I can't implement a better solution for this
# because the TortillaDataFrame constructor enter in conflict 
# with the geopandas constructor.
_original_concat = pd.concat
_original_merge = pd.merge


# Custom pd.concat to handle TortillaDataFrame
# Custom pd.concat to handle TortillaDataFrame
def tortilla_concat(objs, *args, **kwargs):
    if all(isinstance(obj, TortillaDataFrame) for obj in objs):
        result = pd.concat(objs, *args, **kwargs)
        return result.__finalize__(objs[0])
    return pd.concat(objs, *args, **kwargs)

# Monkey patch the pd.concat function
def custom_concat(objs, *args, **kwargs):
    result = _original_concat(objs, *args, **kwargs)

    # Check if any input is a TortillaDataFrame
    if any(isinstance(obj, TortillaDataFrame) for obj in objs):
        result = TortillaDataFrame(gpd.GeoDataFrame(result))
    return result

def custom_merge(left, right, *args, **kwargs):
    result = _original_merge(left, right, *args, **kwargs)

    # Check if any input is a TortillaDataFrame
    if isinstance(left, TortillaDataFrame) or isinstance(right, TortillaDataFrame):
        result = TortillaDataFrame(gpd.GeoDataFrame(result))
    return result

pd.concat = custom_concat   
pd.merge = custom_merge
